//------------------------------------------------------------------------------
// CYmDecruncher.pas
//------------------------------------------------------------------------------
//
// Description: 	Lh5-Decruncher based on a snippet from SWAG archive
//
// Autor:			Sledge Hammer of Grinder
//
// Release:		January 2002
//
// This code is based on the AY_EMUL from Sergey Bulba.
// Read YmPlayer.txt in the docs folder for more information.

//------------------------------------------------------------------------------
// Constructor and Destructor
//------------------------------------------------------------------------------

// Constructor
//
constructor CYmDecruncher.Create;
begin
	m_decrunchedData := nil;
end;

// Destructor
//
destructor CYmDecruncher.Destroy;
begin
	try
       // Free the memory if needed
       if (m_decrunchedData <> nil) then
   		FreeMem(m_decrunchedData);
	finally
  		inherited;
  	end;
end;

//------------------------------------------------------------------------------
// Public declarations
//------------------------------------------------------------------------------

// Loads the ym music from an external file
//
function CYmDecruncher.LoadFromFile(theFilename: string): boolean;
begin
	result := False;

  	// Check if file exists
  	if not FileExists(theFilename) then
   	exit;

  	// Decompress file directly
  	m_decodeFromFile := true;
  	m_badResult := false;
  	m_compressedSize := 0;
	m_originalSize := 0;

  	// Clear the header
   FillChar(m_header, SizeOf(LZHHeader), 0);

   AssignFile(m_inputFile, theFilename);
  	Reset(m_inputFile, 1);
  	BlockRead(m_inputFile, m_header, SizeOf(LZHHeader));

  	// Check for valid compression
  	if (m_header.Method <> '-lh5-') then
   	exit;

  	Seek(m_inputFile, m_header.HSize + 2);
  	m_compressedSize := m_header.CompSize;
  	m_originalSize := m_header.UCompSize;

  	// Free the memory
  	FreeMem(m_decrunchedData);

  	// Get new memory
  	GetMem(m_decrunchedData, m_header.UCompSize);
  	decodeToMemory;

  	CloseFile(m_inputFile);

  	// Free the memory if somethings gone wrong
  	if m_badResult then
   begin
   	FreeMem(m_decrunchedData);
   	m_decrunchedData := nil;
	end;

  	result := not m_badResult;
end;


// Load the ym music from an internal ressource file
//
function CYmDecruncher.LoadFromRessource(theFilename: string): boolean;
var
   HResInfo: HRSRC;
   HGlobal: THandle;
   buffer: pchar;
   ext:string;
begin
 	// Decompress from memory
  	m_decodeFromFile := false;
   m_badResult := true;
  	m_compressedSize := 0;
 	m_originalSize := 0;

 	ext := UpperCase(extractfileext(theFilename));
  	ext:= Copy(ext,2,Length(ext));

  	theFilename:= ChangeFileExt(theFilename,'');
  	HResInfo := FindResource(HInstance, pchar(theFilename), PChar(ext));
  	HGlobal := LoadResource(HInstance, HResInfo);
	if HGlobal = 0 then
   begin
       // Error!
   	result := false;
       exit;
   end;

  	buffer := LockResource(HGlobal);

	// Get the header
  	move(buffer^, m_header, SizeOf(LZHHeader));
  	m_compressedSize := m_header.CompSize;
  	m_originalSize := m_header.UCompSize;

   FreeMem(m_decrunchedData);

	// Get new memory
  	GetMem(m_decrunchedData, m_header.UCompSize);

  	// Move the data the decruch buffer
	m_pointerToBuffer := @buffer^;

  	m_positionInExternalBuf := m_header.HSize + 2;
	m_compressedSizeCopy := m_CompressedSize;
  	m_decodeFromFile := false;

	decodeToMemory;

  	UnlockResource(HGlobal);
  	FreeResource(HGlobal);

  	// Free the memory if somethings gone wrong
  	if m_badResult then
   begin
   	FreeMem(m_decrunchedData);
       m_decrunchedData := nil;
   end;

	result := not m_badResult;
end;


//------------------------------------------------------------------------------
// Private declarations
//------------------------------------------------------------------------------

function CYmDecruncher.getC: Byte;
var
  readen: integer;
begin
  if m_bufPtr = 0 then
	   if m_decodeFromFile then
			   BlockRead(m_inputFile, m_buffer, BUFSIZE, readen)
	   else
	   begin
		    if m_positionInExternalBuf + BUFSIZE > m_compressedSizeCopy then
					   readen := m_compressedSizeCopy - m_positionInExternalBuf
		    else
					   readen := BUFSIZE;
		    move(m_pointerToBuffer^[m_positionInExternalBuf], m_buffer, readen);
		    Inc(m_positionInExternalBuf, readen);
	   end;
  getC := m_buffer[m_bufPtr];
  m_bufPtr := Succ(m_bufPtr) and Pred(BUFSIZE);
end;

function CYmDecruncher.bRead(p: pointer; n: integer): smallint;
begin
  BlockRead(m_inputFile, p^, n, n);
  bRead := n;
end;

function CYmDecruncher.getFromBuffer(from: PVeryLargeBuffer; p: pointer; n: integer): smallint;
begin
  if m_positionInExternalBuf + n > m_originalSize then
	   n := m_originalSize - m_positionInExternalBuf;
  move(from^[m_positionInExternalBuf], p^, n);
  Inc(m_positionInExternalBuf, n);
  getFromBuffer := n;
end;

procedure CYmDecruncher.bwriteToMemory(p: pointer; n: smallint);
begin
  Move(p^, m_decrunchedData^[m_indexInMemory], n);
  inc(m_indexInMemory, n);
end;

procedure CYmDecruncher.fillBuf(n: smallint);
begin
  m_bitBuf := (m_bitBuf shl n);
  while n > m_bitCount do
  begin
	   dec(n, m_bitCount);
	   m_bitBuf := m_bitBuf or (m_subBitBuf shl n);
	   if (m_compressedSize <> 0) then
	   begin
			   dec(m_compressedSize);
		    m_subBitBuf := getC;
	   end
	   else
			   m_subBitBuf := 0;
	   m_bitCount := 8;
  end;
  dec(m_bitCount, n);
  m_bitBuf := m_bitBuf or (m_subBitBuf shr m_bitCount);
end;

function CYmDecruncher.getBits(n: smallint): Word;
begin
  getBits := m_bitBuf shr (BITBUFSIZ - n);
  fillBuf(n);
end;


procedure CYmDecruncher.initGetBits;
begin
  m_bitBuf := 0;
  m_subBitBuf := 0;
  m_bitCount := 0;
  fillBuf(BITBUFSIZ);
end;

procedure CYmDecruncher.makeTable(nchar: smallint; BitLen: PByte;
  TableBits: smallint; Table: PWord);
var
  Count, weight: array[1..16] of Word;
  start: array[1..17] of Word;
  p: PWord;
  i, k, Len, ch, jutbits, Avail, nextCode, mask:{smallint}word;
begin
  for i := 1 to 16 do
	   Count[i] := 0;
  for i := 0 to Pred(nchar) do
	   inc(Count[BitLen^[i]]);
  start[1] := 0;
  for i := 1 to 16 do
	   start[Succ(i)] := start[i] + (Count[i] shl (16 - i));
  if start[17] <> 0 then
  begin
	   {HALT(1);}m_badResult := True;
	   exit
  end;
  jutbits := 16 - TableBits;
  for i := 1 to TableBits do
	 begin
	   start[i] := start[i] shr jutbits;
	   weight[i] := 1 shl (TableBits - i);
	 end;
  i := Succ(TableBits);
  while (i <= 16) do 
  begin
	   weight[i] := 1 shl (16 - i);
	   inc(i);
  end;
  i := start[Succ(TableBits)] shr jutbits;
  if i <> 0 then
	 begin
	   k := 1 shl TableBits;
	   while i <> k do
	   begin
			   Table^[i] := 0;
		    inc(i);
	   end;
	 end;
  Avail := nchar;
  mask := 1 shl (15 - TableBits);
  for ch := 0 to Pred(nchar) do
	 begin
	   Len := BitLen^[ch];
	   if Len = 0 then
			   continue;
	   k := start[Len];
	   nextCode := k + weight[Len];
	   if Len <= TableBits then
		  begin
			   for i := k to Pred(nextCode) do
					   Table^[i] := ch;
		  end 
	   else
	   begin
			   p := Addr(Table^[k shr jutbits]);
		    i := Len - TableBits;
			   while i <> 0 do 
		    begin
					   if p^[0] = 0 then
					   begin
						    m_right[Avail] := 0;
				      m_left[Avail] := 0;
				      p^[0] := Avail;
				      inc(Avail);
					   end;
					   if (k and mask) <> 0 then
					     p := addr(m_right[p^[0]])
					   else
					     p := addr(m_left[p^[0]]);
					   k := k shl 1;
			     dec(i);
			   end;
			   p^[0] := ch;
		  end;
	   start[Len] := nextCode;
	 end;
end;

procedure CYmDecruncher.readPtLen(nn, nBit, ispecial: smallint);
var
  i, c, n: smallint;
  mask: Word;
begin
  n := getBits(nBit);
  if n = 0 then
	 begin
	   c := getBits(nBit);
	   for i := 0 to Pred(nn) do
			   m_ptLen[i] := 0;
	   for i := 0 to 255 do
			   m_ptTable[i] := c;
	 end
  else
  begin
	   i := 0;
	   while (i < n) do
	   begin
			   c := m_bitBuf shr (BITBUFSIZ - 3);
			   if c = 7 then
			   begin
					   mask := 1 shl (BITBUFSIZ - 4);
					   while (mask and m_bitBuf) <> 0 do 
			     begin
					     mask := mask shr 1;
				      inc(c);
					   end;
			   end;
			   if c < 7 then
				    fillBuf(3)
			   else
				    fillBuf(c - 3);
			   m_ptLen[i] := c;
		    inc(i);
			   if i = ispecial then
			   begin
					   c := Pred(getBits(2));
					   while c >= 0 do
			     begin
					     m_ptLen[i] := 0;
				      inc(i);
				      dec(c);
					   end;
			   end;
	   end;
	   while i < nn do
	   begin
			   m_ptLen[i] := 0;
		    inc(i);
	   end;
	   makeTable(nn, @m_ptLen, 8, @m_ptTable);
	 end;
end;

procedure CYmDecruncher.readCLen;
var
  i, c, n: smallint;
  mask: Word;
begin
  n := getBits(CBIT);
  if n = 0 then
	 begin
	   c := getBits(CBIT);
	   for i := 0 to Pred(NC) do
			   m_clen[i] := 0;
	   for i := 0 to 4095 do
			   m_ctable[i] := c;
	 end
  else
  begin
	   i := 0;
	   while i < n do 
	   begin
			   c := m_ptTable[m_bitBuf shr (BITBUFSIZ - 8)];
			   if c >= NT then
			   begin
					   mask := 1 shl (BITBUFSIZ - 9);
					   repeat
					     if (m_bitBuf and mask) <> 0 then
							     c := m_right[c]
					     else
							     c := m_left[c];
					     mask := mask shr 1;
					   until c < NT;
			   end;
			   fillBuf(m_ptLen[c]);
			   if c <= 2 then
			   begin
					   if c = 1 then
					     c := 2 + getBits(4)
					   else if c = 2 then
						    c := 19 + getBits(CBIT);
					   while c >= 0 do
			     begin
					     m_clen[i] := 0;
				      inc(i);
				      dec(c);
					   end;
			   end 
		    else 
		    begin
					   m_clen[i] := c - 2;
			     inc(i);
			   end;
	   end;
	   while i < NC do 
	   begin
			   m_clen[i] := 0;
		    inc(i);
	   end;
	   makeTable(NC, @m_clen, 12, @m_ctable);
	 end;
end;

function CYmDecruncher.decodeC: Word;
var
  j, mask: Word;
begin
  decodeC := 0;
  if m_blockSize = 0 then
	 begin
	   m_blockSize := getBits(16);
	   readPtLen(NT, TBIT, 3);
	   if m_badResult then exit;
	   readCLen;
	   if m_badResult then exit;
	   readPtLen(NP, PBIT, - 1);
	   if m_badResult then exit;
	 end;
  dec(m_blockSize);
  j := m_ctable[m_bitBuf shr (BITBUFSIZ - 12)];
  if j >= NC then
	 begin
	   mask := 1 shl (BITBUFSIZ - 13);
	   repeat
			   if (m_bitBuf and mask) <> 0 then
				    j := m_right[j]
			   else
				    j := m_left[j];
			   mask := mask shr 1;
	   until j < NC;
	 end;
  fillBuf(m_clen[j]);
  decodeC := j;
end;

function CYmDecruncher.decodeP: Word;
var
  j, mask: Word;
begin
  j := m_ptTable[m_bitBuf shr (BITBUFSIZ - 8)];
  if j >= NP then
	 begin
	   mask := 1 shl (BITBUFSIZ - 9);
	   repeat
			   if (m_bitBuf and mask) <> 0 then
				    j := m_right[j]
			   else
				    j := m_left[j];
			   mask := mask shr 1;
	   until j < NP;
	 end;
  fillBuf(m_ptLen[j]);
  if j <> 0 then
	 begin
	   dec(j);
	   j := (1 shl j) + getBits(j);
	 end;
  decodeP := j;
end;

procedure CYmDecruncher.decodeBuffer(Count: Word; Buffer: PByte);
var
  c, r: Word;
begin
  r := 0;
  dec(decode_j);
  while (decode_j >= 0) do
  begin
	   Buffer^[r] := Buffer^[decode_i];
	   decode_i := Succ(decode_i) and Pred(DICSIZ);
	   inc(r);
	   if r = Count then
			   exit;
	   dec(decode_j);
  end;
  while True do
  begin
	   c := decodeC;
	   if m_badResult then exit;
	   if c <= UCHARMAX then
	   begin
			   Buffer^[r] := c;
		    inc(r);
			   if r = Count then
				    exit;
	   end
	   else
	   begin
			   decode_j := c - (UCHARMAX + 1 - THRESHOLD);
			   decode_i := (r - decodeP - 1) and Pred(DICSIZ);
			   dec(decode_j);
			   while decode_j >= 0 do 
		    begin
				    Buffer^[r] := Buffer^[decode_i];
				    decode_i := Succ(decode_i) and Pred(DICSIZ);
				    inc(r);
				    if r = Count then
					     exit;
				    dec(decode_j);
			   end;
	   end;
  end;
end;


procedure CYmDecruncher.decodeToMemory;
var
  p: PByte;
  l: Longint;
  a: Word;
begin
  m_badResult := False;
  m_bufPtr := 0;
  m_indexInMemory := 0;

  // Initialize decoder variables
  GetMem(p, DICSIZ);
  initGetBits;
  m_blockSize := 0;
  decode_j := 0;
  l := m_originalSize;
  while l > 0 do
  begin
	   if l > DICSIZ then
			   a := DICSIZ
	   else
			   a := l;
	   decodeBuffer(a, p);
	   if m_badResult then break;
	   bwriteToMemory(p, a);
	   dec(l, a);
  end;
  FreeMem(p, DICSIZ);
end;




