//------------------------------------------------------------------------------
// CYmGlobals.pas
//------------------------------------------------------------------------------
//
// Description: 	Global constants, types and functions used by the player
//
// Autor:			Sledge Hammer of Grinder
//
// Release:		January 2002
//
// This code is based on the AY_EMUL from Sergey Bulba.
// Read YmPlayer.txt in the docs folder for more information.

unit CYmGlobals;

interface

uses MMSystem,Windows;

//------------------------------------------------------------------------------
// Global types
//------------------------------------------------------------------------------

type

//------------------------------------------------------------------------------
// Global types used by CYmDecruncher
//------------------------------------------------------------------------------

	// Defines an LZH-file header used by CYmDecruncher
	LZHHeader = packed Record
   	HSize 		: byte;
  		CheckSum   	: byte;
  		Method     	: array[1..5] of char;
  		CompSize   	: longint;
  		UCompSize  	: longint;
  		DosDT      	: longint;
  		Attr       	: word;
  		FileNameLen	: byte;
   end;

	VeryLargeBuffer = array[0..2000000000] of byte;
	PVeryLargeBuffer = ^VeryLargeBuffer;
	PWord = ^TWord;
	TWord = array[0..32759] of smallint;
	PByte = ^TByte;
	TByte = array[0..65519] of Byte;

//------------------------------------------------------------------------------
// Global types used by CYmEngine
//------------------------------------------------------------------------------

 	// Defines the ym file types
 	TYmFileTypes = (
 		YM3File,
   	YM3BFile,
   	YM5File,
   	YM6File,
   	Unknown
   	) ;

   // Defines the available soundchips
	TYmChipTypes = (
   	No_Chip,
       YM_Chip
       ) ;

   // Defines the YM5 and YM6 file header structure
 	YM5_Header = packed record
  		Id			: dword;
  		Leo			: array[0..7]of char;
  		Num_of_tiks	: dword;
  		Song_Attr	: dword;
  		Num_of_Dig	: word;
  		ChipFrq		: dword;
  		InterFrq	: word;
  		Loop		: dword;
  		Add_Size	: word;
 	end;

 	// Defines the digital sound data buffer
 	Buf = packed record
 		case boolean of
  			true: (Bit8 : array [1..65280] of byte);
  			false:(Bit16: array [1..32640] of smallint);
 		end;
	PBuf = ^Buf;

 	// Defines the AY register structure
 	TRegisterAY = packed record
 	case Integer of
  		0: (Index: array [0..15] of byte);
  		1: (TonA, TonB, TonC: word;
       	Noise: byte;
     		Mixer: byte;
     		AmplitudeA, AmplitudeB, AmplitudeC: byte;
     		Envelope: word;
     		EnvType: byte);
 	end;

 	// Pointers and stuff
 	BytePtr = ^byte;
 	WordPointer = ^word;
 	DWordPointer = ^longword;
 	ArOfByte = array [0..$FFFFF] of byte;
 	APointer = ^ArOfByte;
	DWordPtr = ^longword;
 	WordPtr = ^word;

//------------------------------------------------------------------------------
// Global constants
//------------------------------------------------------------------------------

const

   // Global constants used by CYmDecruncher
	BITBUFSIZ	= 16;
	UCHARMAX	= 255;
	DICBIT		= 13;
	DICSIZ		= 1 shl DICBIT;
	MATCHBIT	= 8;
	MAXMATCH   	= 1 shl MATCHBIT;
	THRESHOLD	= 3;
	NC			= (UCHARMAX + MAXMATCH + 2 - THRESHOLD);
	CBIT		= 9;
	CODEBIT		= 16;
	NP			= DICBIT+1;
	NT			= CODEBIT+3;
	PBIT		= 4;
	TBIT		= 5;
	NPT			= NT;
	BUFBIT		= 13;
   BUFSIZE		= 1 shl BUFBIT;

   // Default mixer parameters
   SampleRateDef 		= 44100;
 	SampleBitDef  		= 16;
 	FrqZ80Def     		= 3494400;
 	FrqAYDef      		= 1773400;
 	FrqSTDef	  		= 2000000;
 	MaxTStatesDef 		= 69888;
 	Interrupt_FreqDef 	= 50000;
 	Index_ALDef   		= 255;
 	Index_ARDef   		= 13;
 	Index_BLDef   		= 170;
 	Index_BRDef   		= 170;
 	Index_CLDef   		= 13;
 	Index_CRDef   		= 255;
 	NumOfChanDef  		= 2;
 	ChanModeDef   		= 1;
 	MFPTimerModeDef 	= 0;
 	MFPTimerFrqDef 		= 2457600;

   // Player defaults
   PLDef_Number_Of_Channels = 0;
	PLDef_Channel_Mode = -1;
	PLDef_SoundChip_Frq = -1;
	PLDef_Chip_Type = No_Chip;
	PLDef_Player_Frq = -1;

   // Defines the Yamaha 2149 amplitudes table
   Amplitudes_YM: array[0..31] of Word =
    (0,  0,  190,  286, 375, 470, 560, 664, 866, 1130, 1515, 1803, 2253,
    2848, 3351, 3862, 4844, 6058, 7290, 8559, 10474, 12878, 15297, 17787,
    21500, 26172, 30866, 35676, 42664, 50986, 58842, 65535);

	// Defines the wave file header structure
	WaveFileHeader: record
  		rId:          		array [0..3] of char;
  		rLen:         		longint;
  		wId:          		array [0..3] of char;
  		fId:          		array [0..3] of char;
  		fLen:         		longint;
  		wFormatTag:   		word;
  		nChannels:    		word;
  		nSamplesPerSec:  	longint;
  		nAvgBytesPerSec: 	longint;
  		nBlockAlign:  		word;
  		FormatSpecific:  	word;
  		dId:          		array [0..3] of char;
  		dLen:         		longint;
 	end =
 	(rId:'RIFF'; rLen:0; wId:'WAVE'; fId:'fmt '; fLen:16; wFormatTag:1;
  	nChannels:2; nSamplesPerSec:44100; nAvgBytesPerSec:176400;
  	nBlockAlign:4;FormatSpecific:16; dId:'data'; dlen:0);


// Global variables
var
   Synthesizer: procedure;
   First_Period: boolean;

   Ton_Counter_A, Ton_Counter_B, Ton_Counter_C, Noise_Counter: packed record
   	case integer of
      		0: (Lo: word; Hi: word);
      		1: (Re: longword);
   	end;

   Envelope_Counter: packed record
   	case integer of
     		0: (Lo: dword; Hi: dword);
     		1: (Re: int64);
     	end;

   Ton_A,Ton_B,Ton_C: byte;

	Noise: packed record
   	case boolean of
      		true: (Seed: longword);
      		false:(Low: word; Val: word);
     	end;

   RegisterAY: TRegisterAY;
   Ampl: smallint;
   Level_AR, Level_AL,
   Level_BR, Level_BL,
   Level_CR, Level_CL: array[0..31]of integer;
   Left_Chan, Right_Chan: integer;
   Tick_Counter: byte;

   Tik: packed record
   	case Integer of
      		0: (Lo: word; Hi: word);
      		1: (Re: dword);
     	end;

   Delay_in_tiks: dword;
   BuffLen: word;
   PBuffer: PBuf;
   Current_Tik: longword;

   Number_Of_Tiks: packed record
   	case boolean of
      		false:(lo: longword; hi: longword);
      		true: (re: int64);
     	end;

   MaxBuffLen,MaxSamples: longword;
   IntFlag: boolean;
   Ton_EnA, Ton_EnB, Ton_EnC,
   Noise_EnA, Noise_EnB,Noise_EnC: boolean;
   Envelope_EnA,Envelope_EnB,Envelope_EnC: boolean;
   Case_EnvType: procedure;
   MaxTStates: integer;

   // Global vars used by CYmDecruncher
  	decode_i: word;
  	decode_j: smallint;

//------------------------------------------------------------------------------
// Declaration of global functions used by the engine
//------------------------------------------------------------------------------

   function IntelWord(Wrd:word):word;
	function IntelSmallInt(SmI:smallint):smallint;
	procedure Case_EnvType_0_3__9;
	procedure Case_EnvType_4_7__15;
	procedure Case_EnvType_8;
	procedure Case_EnvType_10;
	procedure Case_EnvType_11;
	procedure Case_EnvType_12;
	procedure Case_EnvType_13;
   procedure Case_EnvType_14;
   procedure SetMixerRegister(Value: byte);
   procedure SetEnvelopeRegister(Value: byte);
   procedure SetAmplA(Value: byte);
   procedure SetAmplB(Value: byte);
   procedure SetAmplC(Value: byte);
   procedure Synthesizer_Logic_Q;
   procedure Synthesizer_Logic_P;
   procedure Synthesizer_Mixer_Q;
   procedure Synthesizer_Stereo16;
   procedure Synthesizer_Stereo8;
   procedure Synthesizer_Mixer_Q_Mono;
   procedure Synthesizer_Mono16;
   procedure Synthesizer_Mono8;
   procedure Synthesizer_Stereo16_P;
   procedure Synthesizer_Stereo8_P;
   procedure Synthesizer_Mono16_P;
   procedure Synthesizer_Mono8_P;

implementation

//------------------------------------------------------------------------------
// Implementation of global functions used by the engine
//------------------------------------------------------------------------------

function IntelWord(Wrd:word):word;
asm
	xchg al,ah
end;

function IntelSmallInt(SmI:smallint):smallint;
asm
	xchg al,ah
end;

procedure Case_EnvType_0_3__9;
begin
	if First_Period then
	begin
		dec(Ampl);
		if Ampl = 0 then First_Period := False
	end
end;

procedure Case_EnvType_4_7__15;
begin
	if First_Period then
	begin
		inc(Ampl);
		if Ampl = 32 then
		begin
			First_Period := False;
			Ampl := 0;
		end
	end
end;

procedure Case_EnvType_8;
begin
	Ampl := (Ampl - 1) and 31;
end;

procedure Case_EnvType_10;
begin
	if First_Period then
	begin
		dec(Ampl);
		if Ampl = -1 then
		begin
			First_Period := False;
			Ampl := 0
		end
	end
	else
	begin
		inc(Ampl);
		if Ampl = 32 then
		begin
			First_Period := True;
			Ampl := 31
		end
	end
end;

procedure Case_EnvType_11;
begin
	if First_Period then
	begin
		dec(Ampl);
		if Ampl = -1 then
		begin
			First_Period := False;
			Ampl := 31
		end
	end
end;

procedure Case_EnvType_12;
begin
	Ampl := (Ampl + 1) and 31;
end;

procedure Case_EnvType_13;
begin
	if First_Period then
	begin
		inc(Ampl);
		if Ampl = 32 then
		begin
			First_Period := False;
			Ampl := 31
		end
	end
end;

procedure Case_EnvType_14;
begin
	if not First_Period then
	begin
		dec(Ampl);
		if Ampl = -1 then
		begin
			First_Period := True;
			Ampl := 0
		end
	end
	else
	begin
		inc(Ampl);
		if Ampl = 32 then
		begin
			First_Period := False;
			Ampl := 31
		end
	end
end;


procedure SetMixerRegister(Value: byte);
begin
	RegisterAY.Mixer := Value;
	Ton_EnA := (Value and 1) = 0;
	Noise_EnA := (Value and 8) = 0;
	Ton_EnB := (Value and 2) = 0;
	Noise_EnB := (Value and 16) = 0;
	Ton_EnC := (Value and 4) = 0;
	Noise_EnC := (Value and 32) = 0
end;

procedure SetEnvelopeRegister(Value: byte);
begin
	Envelope_Counter.Hi := 0;
	First_Period := True;
	if (Value and 4) = 0 then
		ampl := 32
	else
		ampl := -1;
	RegisterAY.EnvType := Value;
	case Value of
		0..3,9: Case_EnvType := Case_EnvType_0_3__9;
		4..7,15: Case_EnvType := Case_EnvType_4_7__15;
		8: Case_EnvType := Case_EnvType_8;
		10: Case_EnvType := Case_EnvType_10;
		11: Case_EnvType := Case_EnvType_11;
		12: Case_EnvType := Case_EnvType_12;
		13: Case_EnvType := Case_EnvType_13;
		14: Case_EnvType := Case_EnvType_14;
	end;
end;

procedure SetAmplA(Value: byte);
begin
	RegisterAY.AmplitudeA := Value;
	Envelope_EnA := (Value and 16) = 0;
end;

procedure SetAmplB(Value: byte);
begin
	RegisterAY.AmplitudeB := Value;
	Envelope_EnB := (Value and 16) = 0;
end;

procedure SetAmplC(Value: byte);
begin
	RegisterAY.AmplitudeC := Value;
	Envelope_EnC := (Value and 16) = 0;
end;

procedure Synthesizer_Logic_Q;
begin
	inc(Ton_Counter_A.Hi);
	if Ton_Counter_A.Hi >= RegisterAY.TonA then
	begin
		Ton_Counter_A.Hi := 0;
		Ton_A := Ton_A xor 1
	end;
	inc(Ton_Counter_B.Hi);
	if Ton_Counter_B.Hi >= RegisterAY.TonB then
	begin
		Ton_Counter_B.Hi := 0;
		Ton_B := Ton_B xor 1
	end;
	inc(Ton_Counter_C.Hi);
	if Ton_Counter_C.Hi >= RegisterAY.TonC then
	begin
		Ton_Counter_C.Hi := 0;
		Ton_C := Ton_C xor 1
	end;
	inc(Noise_Counter.Hi);
	if (Noise_Counter.Hi and 1 = 0) and
		(Noise_Counter.Hi >= RegisterAY.Noise shl 1) then
	begin
		Noise_Counter.Hi := 0;
		Noise.Val := Random(2); {Real noise algorithm is forbidden}
	end;
	if Envelope_Counter.Hi = 0 then Case_EnvType;
	inc(Envelope_Counter.Hi);
	if Envelope_Counter.Hi >= RegisterAY.Envelope then
		Envelope_Counter.Hi := 0
end;

procedure Synthesizer_Logic_P;
var
	k: word;
	k2: longword;
begin
	inc(Ton_Counter_A.Re, Delay_In_Tiks);
	if RegisterAY.TonA = 0 then k := 1
	else
		k := RegisterAY.TonA;
	if k <= Ton_Counter_A.Hi then
	begin
		Ton_A := Ton_A xor ((Ton_Counter_A.Hi div k) and 1);
		Ton_Counter_A.Hi := Ton_Counter_A.Hi mod k
	end;

	inc(Ton_Counter_B.Re, Delay_In_Tiks);
	if RegisterAY.TonB = 0 then k := 1
	else
		k := RegisterAY.TonB;
	if k <= Ton_Counter_B.Hi then
	begin
		Ton_B := Ton_B xor ((Ton_Counter_B.Hi div k) and 1);
		Ton_Counter_B.Hi := Ton_Counter_B.Hi mod k
	end;

	inc(Ton_Counter_C.Re, Delay_In_Tiks);
	if RegisterAY.TonC = 0 then k := 1
	else
		k := RegisterAY.TonC;
	if k <= Ton_Counter_C.Hi then
	begin
		Ton_C := Ton_C xor ((Ton_Counter_C.Hi div k) and 1);
		Ton_Counter_C.Hi := Ton_Counter_C.Hi mod k
	end;

	inc(Noise_Counter.Re, Delay_In_Tiks);
	if RegisterAY.Noise = 0 then k := 2
	else
		k := RegisterAY.Noise shl 1;
	if Noise_Counter.Hi >= k then
	begin
		Noise_Counter.Hi := Noise_Counter.Hi mod k;
		Noise.Val := Random(2); {Real noise algorithm is forbidden}
	end;

	if RegisterAY.Envelope = 0 then k2 := 1
	else
		k2 := RegisterAY.Envelope;
	if Envelope_Counter.Hi = 0 then inc(Envelope_Counter.Hi, k2);
	while (Envelope_Counter.Hi >= k2) do
	begin
		dec(Envelope_Counter.Hi, k2);
		Case_EnvType
	end;
	inc(Envelope_Counter.Re, int64(Delay_In_Tiks) shl 16);
end;

procedure Synthesizer_Mixer_Q;
var
	LevL, LevR, k: integer;
begin
	LevL := 0;
	LevR := LevL;
	if Ton_EnA then k := Ton_A
	else
		k := 1;
	if Noise_EnA then k := k and Noise.Val;
	if k <> 0 then
	begin
		if Envelope_EnA then
		begin
			LevL := Level_AL[RegisterAY.AmplitudeA * 2 + 1];
			LevR := Level_AR[RegisterAY.AmplitudeA * 2 + 1];
		end
		else
		begin
			LevL := Level_AL[Ampl];
			LevR := Level_AR[Ampl];
		end
	end;

	if Ton_EnB then k := Ton_B
	else
		k := 1;
	if Noise_EnB then k := k and Noise.Val;
	if k <> 0 then
		if Envelope_EnB then
		begin
			inc(LevL, Level_BL[RegisterAY.AmplitudeB * 2 + 1]);
			inc(LevR, Level_BR[RegisterAY.AmplitudeB * 2 + 1]);
		end
		else
		begin
			inc(LevL, Level_BL[Ampl]);
			inc(LevR, Level_BR[Ampl]);
		end;

	if Ton_EnC then k := Ton_C
	else
		k := 1;
	if Noise_EnC then k := k and Noise.Val;
	if k <> 0 then
		if Envelope_EnC then
		begin
			inc(LevL, Level_CL[RegisterAY.AmplitudeC * 2 + 1]);
			inc(LevR, Level_CR[RegisterAY.AmplitudeC * 2 + 1]);
		end
		else
		begin
			inc(LevL, Level_CL[Ampl]);
			inc(LevR, Level_CR[Ampl]);
		end;
	inc(Left_Chan, LevL);
	inc(Right_Chan, LevR)
end;


procedure Synthesizer_Stereo16;
var
	tmp: integer;
begin
	repeat
		Synthesizer_Logic_Q;
		Synthesizer_Mixer_Q;
		inc(Current_Tik);
		inc(Tick_Counter);
		if Tick_Counter >= Tik.Hi then
		begin
			inc(Tik.Re, Delay_In_Tiks);
			dec(Tik.Hi, Tick_Counter);
			inc(BuffLen);
			PBuffer^.Bit16[BuffLen] := Left_Chan div Tick_Counter;
			inc(BuffLen);
			PBuffer^.Bit16[BuffLen] := Right_Chan div Tick_Counter;
			Tmp := 0;
			Left_Chan := Tmp;
			Right_Chan := Tmp;
			Tick_Counter := Tmp;
			if BuffLen = MaxBuffLen then
			begin
				if Current_Tik < Number_Of_Tiks.Hi then
					IntFlag := True;
				exit
			end
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	Tmp := 0;
	Number_Of_Tiks.hi := Tmp;
	Current_Tik := Tmp
end;


procedure Synthesizer_Stereo8;
var
	Tmp: integer;
begin
	repeat
		Synthesizer_Logic_Q;
		Synthesizer_Mixer_Q;
		inc(Current_Tik);
		inc(Tick_Counter);
		if Tick_Counter >= Tik.Hi then
		begin
			inc(Tik.Re, Delay_In_Tiks);
			dec(Tik.Hi, Tick_Counter);
			inc(BuffLen);
			PBuffer^.Bit8[BuffLen] := 128 + Left_Chan div Tick_Counter;
			inc(BuffLen);
			PBuffer^.Bit8[BuffLen] := 128 + Right_Chan div Tick_Counter;
			Tmp := 0;
			Left_Chan := Tmp;
			Right_Chan := Tmp;
			Tick_Counter := Tmp;
			if BuffLen = MaxBuffLen then
			begin
				if Current_Tik < Number_Of_Tiks.Hi then
					IntFlag := True;
				exit
			end
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	Tmp := 0;
	Number_Of_Tiks.hi := Tmp;
	Current_Tik := Tmp
end;


procedure Synthesizer_Mixer_Q_Mono;
var
	Lev, k: integer;
begin
	Lev := 0;
	if Ton_EnA then k := Ton_A
	else
		k := 1;
	if Noise_EnA then k := k and Noise.Val;
	if k <> 0 then
		if Envelope_EnA then
			Lev := Level_AL[RegisterAY.AmplitudeA * 2 + 1]
		else
			Lev := Level_AL[Ampl];

	if Ton_EnB then k := Ton_B
	else
		k := 1;
	if Noise_EnB then k := k and Noise.Val;
	if k <> 0 then
		if Envelope_EnB then
			inc(Lev, Level_BL[RegisterAY.AmplitudeB * 2 + 1])
		else
			inc(Lev, Level_BL[Ampl]);

	if Ton_EnC then k := Ton_C
	else
		k := 1;
	if Noise_EnC then k := k and Noise.Val;
	if k <> 0 then
		if Envelope_EnC then
			inc(Lev, Level_CL[RegisterAY.AmplitudeC * 2 + 1])
		else
			inc(Lev, Level_CL[Ampl]);
	inc(Left_Chan, Lev)
end;


procedure Synthesizer_Mono16;
var
	Tmp: integer;
begin
	repeat
		Synthesizer_Logic_Q;
		Synthesizer_Mixer_Q_Mono;
		inc(Current_Tik);
		inc(Tick_Counter);
		if Tick_Counter >= Tik.Hi then
		begin
			inc(Tik.Re, Delay_In_Tiks);
			dec(Tik.Hi, Tick_Counter);
			inc(BuffLen);
			PBuffer^.Bit16[BuffLen] := Left_Chan div Tick_Counter;
			Tmp := 0;
			Left_Chan := Tmp;
			Tick_Counter := Tmp;
			if BuffLen = MaxBuffLen then
			begin
				if Current_Tik < Number_Of_Tiks.Hi then
					IntFlag := True;
				exit
			end
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	Tmp := 0;
	Number_Of_Tiks.hi := Tmp;
	Current_Tik := Tmp
end;

procedure Synthesizer_Mono8;
var
	Tmp: integer;
begin
	repeat
		Synthesizer_Logic_Q;
		Synthesizer_Mixer_Q_Mono;
		inc(Current_Tik);
		inc(Tick_Counter);
		if Tick_Counter >= Tik.Hi then
		begin
			inc(Tik.Re, Delay_In_Tiks);
			dec(Tik.Hi, Tick_Counter);
			inc(BuffLen);
			PBuffer^.Bit8[BuffLen] := 128 + Left_Chan div Tick_Counter;
			Tmp := 0;
			Left_Chan := Tmp;
			Tick_Counter := Tmp;
			if BuffLen = MaxBuffLen then
			begin
				if Current_Tik < Number_Of_Tiks.Hi then
					IntFlag := True;
				exit
			end
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	Tmp := 0;
	Number_Of_Tiks.hi := Tmp;
	Current_Tik := Tmp
end;


procedure Synthesizer_Stereo16_P;
var
	LevL, LevR, k: integer;
begin
	repeat
		Synthesizer_Logic_P;
		LevL := 0;
		LevR := LevL;

		if Ton_EnA then k := Ton_A
		else
			k := 1;
		if Noise_EnA then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnA then
			begin
				LevL := Level_AL[RegisterAY.AmplitudeA * 2 + 1];
				LevR := Level_AR[RegisterAY.AmplitudeA * 2 + 1];
			end
			else
			begin
				LevL := Level_AL[Ampl];
				LevR := Level_AR[Ampl];
			end;

		if Ton_EnB then k := Ton_B
		else
			k := 1;
		if Noise_EnB then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnB then
			begin
				inc(LevL, Level_BL[RegisterAY.AmplitudeB * 2 + 1]);
				inc(LevR, Level_BR[RegisterAY.AmplitudeB * 2 + 1]);
			end
			else
			begin
				inc(LevL, Level_BL[Ampl]);
				inc(LevR, Level_BR[Ampl]);
			end;

		if Ton_EnC then k := Ton_C
		else
			k := 1;
		if Noise_EnC then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnC then
			begin
				inc(LevL, Level_CL[RegisterAY.AmplitudeC * 2 + 1]);
				inc(LevR, Level_CR[RegisterAY.AmplitudeC * 2 + 1]);
			end
			else
			begin
				inc(LevL, Level_CL[Ampl]);
				inc(LevR, Level_CR[Ampl]);
			end;

		inc(BuffLen);
		PBuffer^.Bit16[BuffLen] := LevL;
		inc(BuffLen);
		PBuffer^.Bit16[BuffLen] := LevR;
		inc(Current_Tik);
		if BuffLen = MaxBuffLen then
		begin
			if Current_Tik < Number_Of_Tiks.Hi then
				IntFlag := True;
			exit
		end;
	until Current_Tik >= Number_Of_Tiks.Hi;
	k := 0;
	Number_Of_Tiks.hi := k;
	Current_Tik := k
end;


procedure Synthesizer_Stereo8_P;
var
	LevL, LevR, k: integer;
begin
	repeat
		Synthesizer_Logic_P;
		LevL := 128;
		LevR := LevL;

		if Ton_EnA then k := Ton_A
		else
			k := 1;
		if Noise_EnA then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnA then
			begin
				inc(LevL, Level_AL[RegisterAY.AmplitudeA * 2 + 1]);
				inc(LevR, Level_AR[RegisterAY.AmplitudeA * 2 + 1]);
			end
			else
			begin
				inc(LevL, Level_AL[Ampl]);
				inc(LevR, Level_AR[Ampl]);
			end;

		if Ton_EnB then k := Ton_B
		else
			k := 1;
		if Noise_EnB then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnB then
			begin
				inc(LevL, Level_BL[RegisterAY.AmplitudeB * 2 + 1]);
				inc(LevR, Level_BR[RegisterAY.AmplitudeB * 2 + 1]);
			end
			else
			begin
				inc(LevL, Level_BL[Ampl]);
				inc(LevR, Level_BR[Ampl]);
			end;

		if Ton_EnC then k := Ton_C
		else
			k := 1;
		if Noise_EnC then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnC then
			begin
				inc(LevL, Level_CL[RegisterAY.AmplitudeC * 2 + 1]);
				inc(LevR, Level_CR[RegisterAY.AmplitudeC * 2 + 1]);
			end
			else
			begin
				inc(LevL, Level_CL[Ampl]);
				inc(LevR, Level_CR[Ampl]);
			end;

		inc(BuffLen);
		PBuffer^.Bit8[BuffLen] := LevL;
		inc(BuffLen);
		PBuffer^.Bit8[BuffLen] := LevR;
		inc(Current_Tik);
		if BuffLen = MaxBuffLen then
		begin
			if Current_Tik < Number_Of_Tiks.Hi then
				IntFlag := True;
			exit
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	k := 0;
	Number_Of_Tiks.hi := k;
	Current_Tik := k
end;


procedure Synthesizer_Mono16_P;
var
	Lev, k: integer;
begin
	repeat
		Synthesizer_Logic_P;
		Lev := 0;

		if Ton_EnA then k := Ton_A
		else
			k := 1;
		if Noise_EnA then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnA then
				Lev := Level_AL[RegisterAY.AmplitudeA * 2 + 1]
			else
				Lev := Level_AL[Ampl];

		if Ton_EnB then k := Ton_B
		else
			k := 1;
		if Noise_EnB then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnB then
				inc(Lev, Level_BL[RegisterAY.AmplitudeB * 2 + 1])
			else
				inc(Lev, Level_BL[Ampl]);

		if Ton_EnC then k := Ton_C
		else
			k := 1;
		if Noise_EnC then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnC then
				inc(Lev, Level_CL[RegisterAY.AmplitudeC * 2 + 1])
			else
				inc(Lev, Level_CL[Ampl]);

		inc(BuffLen);
		PBuffer^.Bit16[BuffLen] := Lev;
		inc(Current_Tik);
		if BuffLen = MaxBuffLen then
		begin
			if Current_Tik < Number_Of_Tiks.Hi then
				IntFlag := True;
			exit
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	k := 0;
	Number_Of_Tiks.hi := k;
	Current_Tik := k
end;


procedure Synthesizer_Mono8_P;
var
	Lev, k: integer;
begin
	repeat
		Synthesizer_Logic_P;
		Lev := 128;

		if Ton_EnA then k := Ton_A
		else
			k := 1;
		if Noise_EnA then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnA then
				inc(Lev, Level_AL[RegisterAY.AmplitudeA * 2 + 1])
			else
				inc(Lev, Level_AL[Ampl]);

		if Ton_EnB then k := Ton_B
		else
			k := 1;
		if Noise_EnB then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnB then
				inc(Lev, Level_BL[RegisterAY.AmplitudeB * 2 + 1])
			else
				inc(Lev, Level_BL[Ampl]);

		if Ton_EnC then k := Ton_C
		else
			k := 1;
		if Noise_EnC then k := k and Noise.Val;
		if k <> 0 then
			if Envelope_EnC then
				inc(Lev, Level_CL[RegisterAY.AmplitudeC * 2 + 1])
			else
				inc(Lev, Level_CL[Ampl]);

		inc(BuffLen);
		PBuffer^.Bit8[BuffLen] := Lev;
		inc(Current_Tik);
		if BuffLen = MaxBuffLen then
		begin
			if Current_Tik < Number_Of_Tiks.Hi then
				IntFlag := True;
			exit
		end
	until Current_Tik >= Number_Of_Tiks.Hi;
	k := 0;
	Number_Of_Tiks.hi := k;
	Current_Tik := k
end;

end.
