//------------------------------------------------------------------------------
// CYmPlayer.pas
//------------------------------------------------------------------------------
//
// Description: 	Main player classes
//
// Autor:			Sledge Hammer of Grinder
//
// Release:		January 2002
//
// This code is based on the AY_EMUL from Sergey Bulba.
// Read YmPlayer.txt in the docs folder for more information.

unit CYmPlayer;

interface

uses MMSystem, SysUtils, Windows, CYmGlobals;

type
   // Defines the optimization mode
   TYmOptimizationMode = (
   	forPerformance,
   	forQuality
   	) ;

	// Defines the waveout channels
   TYmSampleChannels = (
		channelsMono,
   	channelsStereo
   	) ;

   // Defines the waveout sample bits
	TYmSampleBits = (
		bits8,
   	bits16
   	) ;

   // Defines the waveout sample frequency
	TYmSampleFrequency = (
		freq11025,
   	freq22050,
   	freq44100
   	) ;

   // Defines the ym mixer channel mode
	TYmMixerMode = (
   	modeMono,
   	modeABC,
   	modeACB,
   	modeBAC,
   	modeBCA,
   	modeCAB,
   	modeCBA,
   	modeManual
		) ;

	// CYmSong encapsulates the Ym song item
	CYmSong = class(TObject)
       // Public declarations
  		public
  			Author: string;
           Mod_Name: string;
           FileName: string;
           Prog_Name: string;
           Track_Name: string;
           Comp_Name: string;
  			Date_String: string;
           Comment_String: string;
  			Time_Length: integer;
           Offset: integer;
           Loop_VBL: integer;
  			Address: integer;
           Length: integer;
  			FileType: TYmFileTypes;
  			USize: integer;
  			SoundChip_Frq: integer;
  			Player_Frq: integer;
  			Channel_Mode: TYmMixerMode;
  			Chip_Type: TYmChipTypes;
  			Number_Of_Channels: integer;
  			AL: byte;
           AR: byte;
           BL: byte;
           BR: byte;
           CL: byte;
           CR: byte;
  		public
  			constructor Create;
			destructor Destroy; override;
  		public
  			procedure Reset;
  	end;

   	// CYmDecruncher encapsulates the internal file decruncher
	CYmDecruncher = class
   	// Private declarations
  		private
	   		m_inputFile: file;
	   		m_header: LZHHeader;
	   		m_indexInMemory: integer;
	   		m_positionInExternalBuf: longword;
	   		m_compressedSizeCopy: longword;
	   		m_bitBuf: Word;
	   		m_subBitBuf: word;
      		m_bitCount: Word;
	   		m_buffer: array[0..Pred(BUFSIZE)] of byte;
	   		m_bufPtr: Word;
	   		m_left:  array[0..2 * (NC - 1)] of word;
      		m_right: array[0..2 * (NC - 1)] of word;
	   		m_ptTable: array[0..255] of Word;
	   		m_ptLen: array[0..Pred(NPT)] of Byte;
	   		m_ctable: array[0..4095] of Word;
	   		m_clen: array[0..Pred(NC)] of Byte;
	   		m_blockSize: Word;
	   		m_decrunchedData: PVeryLargeBuffer;
	   		m_pointerToBuffer: PVeryLargeBuffer;
	   		m_badResult: boolean;
	   		m_decodeFromFile: boolean;
	   		m_compressedSize: integer;
      		m_originalSize: integer;
  		private
	   		function getC: Byte;
	   		function bRead(p: pointer; n: integer): smallint;
	   		function getFromBuffer(from: PVeryLargeBuffer; p: pointer; n: integer): smallint;
	   		procedure bwriteToMemory(p: pointer; n: smallint);
	   		procedure fillBuf(n: smallint);
	   		function getBits(n: smallint): Word;
	   		procedure initGetBits;
	   		procedure makeTable(nchar: smallint; BitLen: PByte; TableBits: smallint; Table: PWord);
	   		procedure readPtLen(nn, nBit, ispecial: smallint);
	   		procedure readCLen;
	   		function decodeC: Word;
	   		function decodeP: Word;
	   		procedure decodeBuffer(Count: Word; Buffer: PByte);
			procedure decodeToMemory;
       // Public declarations
  		public
	 	 	constructor Create;
		  	destructor Destroy; override;
  		public
	   		function LoadFromFile(theFilename: string): boolean;
      		function LoadFromRessource(theFilename: string): boolean;
  		public
	   		property Header: LZHHeader read m_header;
	   		property DecrunchedData: PVeryLargeBuffer read m_decrunchedData;
	   		property OriginalSize: integer read m_originalSize;
	   		property CompressedSize: integer read m_compressedSize;
   end;

  	// Encapsulates the main ym player engine
	CYmEngine = class
   	// Private declarations
       private
       	m_ymSong: CYmSong;
			m_ymDecruncher: CYmDecruncher;
       	m_optimizationForQuality: boolean;
			m_isLoaded: boolean;
       	m_channelMode: TYmMixerMode;
			m_phwo: hwaveout;
			SampleRate: word;
			FrqZ80: longword;
			FrqAY: longword;
			m_isPlaying: boolean;
			Reseted: boolean;
			Real_End: boolean;
			Buffer1, Buffer2: PBuf;
			Current_RegisterAY: byte;
			pwh1, pwh2: wavehdr;
			m_isPaused: boolean;
			Time_Rasch, BaseTime: longint;
			FileOutDataCounter: integer;
			FileType: TYmFileTypes;
			ProgrMax: longword;
			Tiks_In_Interrupt: word;
			Sample_Tiks_in_Interrupt: word;
			Interrupt_Freq: longword;
			YM5Header: YM5_Header;
			Position_In_VTX: longword;
			Number_Of_Tiks_In_VTX: longword;
			VTX_Offset: dword;
			may_quit: boolean;
			VBL_For_loop: dword;
			m_PlayInLoop: boolean;
			Emulating_Chip: TYmChipTypes;
			Index_AL, Index_AR, Index_BL, Index_BR, Index_CL, Index_CR: byte;
			SampleBit: byte;
			NumOfChan: integer;
			Global_Tick_Counter: longword;
			Global_Tick_Max: longword;
			m_preAmp, PreAmpMax: byte;
			DDrumSamples: array of record
				Length: integer;
				Buf: APointer;
			end;
			AtariSE1Channel, AtariSE2Channel, AtariSE1TP, AtariSE2TP: byte;
			AtariSE1Type, AtariSE2Type: byte;
			AtariSE1Pos, AtariSE2Pos: integer;
			AtariParam1, AtariParam2: byte;
			AtariV1, AtariV2: byte;
			AtariTimerCounter1, AtariTimerCounter2: real;
			AtariTimerPeriod1, AtariTimerPeriod2: real;
			YM6TiksOnInt, YM6CurTik: real;
			YM6SamTiksOnInt: real;
			YM6SinusTable: array [0..15,0..7] of byte;
			YM6SinusPos1, YM6SinusPos2: integer;
			YM6Tiks: int64;
			MFPTimerFrq, MFPTimerMode: integer;
			MakeBuffer: procedure of object;
       private
       	procedure initDefaults;
			procedure getLength;
			procedure makeBufferVTX;
			procedure makeBufferYM5;
			procedure makeBufferYM6;
			procedure synthesizerZX50;
			procedure synthesizerYM6;
			procedure InitForAllTypes;
			procedure vtxYm3bGetRegisters;
			procedure ym5GetRegisters;
			procedure ym6GetRegisters;
			procedure ym5iGetRegisters;
			procedure ym6iGetRegisters;
			procedure ym6GetExtraRegisters;
			function revers(a: integer; b: integer): dword;
			procedure raiseWaveOutError(Text: PChar; Title: string);
			procedure calculateLevelTables;

			procedure setOptimizationFlag(Q: boolean);
			procedure resetAYChipEmulation;
           procedure getYmInformation;
       private
      	function getSampleRate: TYmSampleFrequency;
       	procedure setSampleRate(theRate: TYmSampleFrequency);
           function getMixerMode: TYmMixerMode;
			procedure setMixerMode(theMixerMode: TYmMixerMode);
           function getOutputChannels: TYmSampleChannels;
			procedure setOutputChannels(theChannels: TYmSampleChannels);
			function getOutputBits: TYmSampleBits;
			procedure setOutputBits(theBits: TYmSampleBits);
           procedure setOptimization(theOptimization: TYmOptimizationMode);
           function getOptimization: TYmOptimizationMode;
			procedure setPreampVolume(theVolume: byte);
       // Public declarations
       public
       	constructor Create;
           destructor Destroy; override;
       public
            procedure setPlayInLoop(theValue: boolean);
       			procedure setChipFrq(Fr: integer);
       	function LoadFromRessource(theFilename: string): boolean;
			function LoadFromFile(theFilename: string): boolean;
			procedure Play;
			procedure Stop;
			procedure Pause;
           procedure Reset;
           procedure SetMixerModeManual(ALeft, ARight, BLeft, BRight, CLeft, CRight: byte);
       public
       	property IsPlaying: boolean read m_isPlaying;
           property IsPaused: boolean read m_isPaused;
           property PlayInLoop: boolean read m_playInLoop write setPlayInLoop;
           property Volume: byte read m_preAmp write setPreAmpVolume;
   		property MixerChannels: TYmMixerMode read getMixerMode write setMixerMode;
   		property SampleFrequency: TYmSampleFrequency read getSampleRate write setSampleRate;
   		property SampleChannels: TYmSampleChannels read getOutputChannels write setOutputChannels;
   		property SampleBits: TYmSampleBits read getOutputBits write setOutputBits;
   	    property OptimizationMode: TYmOptimizationMode read getOptimization write setOptimization;

  	end;

var
   // Global engine reference for playback callback
	engine: CYmEngine;

implementation

// The wave out playback procedure
//
procedure waveOutProc(hwo: HWAVEOUT; uMsg: UINT; dwInstance, dwParam1, dwParam2: DWORD); stdcall;
var
	Res: Longint;
	ErrMsg: array [0..255] of Char;
begin
	with engine do
	begin
		if (uMsg <> WOM_DONE) or (PWaveHdr(dwParam1).dwUser <> 0) then
       	exit;

		if Reseted or not m_isPlaying then
		begin
			PWaveHdr(dwParam1).dwUser := 1;
			exit;
		end;
		if not Real_End then
		begin
			if dwParam1 = DWORD(@pwh1) then
			begin
				PBuffer := Buffer1;
			end
			else
			begin
				PBuffer := Buffer2;
			end;
			inc(FileOutDataCounter, PWaveHdr(dwParam1).dwBufferLength);
			MakeBuffer;
			if BuffLen <> 0 then
			begin
				PWaveHdr(dwParam1).dwBufferLength := BuffLen * (SampleBit div 8);
				res := waveOutWrite(m_phwo, pointer(dwParam1), SizeOf(WaveHdr));
				if res <> 0 then
					waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			end
			else
				PWaveHdr(dwParam1).dwUser := 1;
		end
		else
		begin
			if (pwh1.dwFlags and whdr_DONE = 0) or (pwh2.dwFlags and whdr_DONE = 0) then
			begin
				if pwh1.dwFlags and whdr_DONE <> 0 then
					pwh1.dwUser := 1;
				if pwh2.dwFlags and whdr_DONE <> 0 then
					pwh2.dwUser := 1;

				inc(FileOutDataCounter, 65280);
				exit;
			end;
			pwh1.dwUser := 1;
			pwh2.dwUser := 1;
			FileOutDataCounter := ProgrMax;
			Stop;
		end
	end
end;

//------------------------------------------------------------------------------
// Implementation of class CYmSong
//------------------------------------------------------------------------------
{$I CYmSong.pas}

//------------------------------------------------------------------------------
// Implementation of class CYmDecruncher
//------------------------------------------------------------------------------
{$I CYmDecruncher.pas}

//------------------------------------------------------------------------------
// Implementation of class CYmEngine
//------------------------------------------------------------------------------
{$I CYmEngine.pas}


end.

